// SPDX-License-Identifier: AGPL-3.0-or-later
// SPDX-FileCopyrightText: 2023 swt23w23
package catering.staff;

import java.time.LocalDateTime;
import java.time.YearMonth;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.salespointframework.order.OrderManagement;
import org.springframework.data.domain.Pageable;
import org.springframework.data.util.Streamable;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.SessionAttributes;

import catering.order.CustomOrder;

@Component
@Transactional
@SessionAttributes("staff")
public class StaffManagement {

	private final StaffRepository staffRepository;
	private final OrderManagement<CustomOrder> orderManagement;

	public StaffManagement(StaffRepository staffRepository, OrderManagement<CustomOrder> orderManagement) {
		this.staffRepository = staffRepository;
		this.orderManagement = orderManagement;
	}

	public Optional<Employee> findById(Long id) {
		return staffRepository.findById(id);
	}

	public Employee save(Employee employee) {
		return staffRepository.save(employee);
	}

	public Streamable<Employee> findAll() {
		return staffRepository.findAll();
	}

	public void delete(Long id) {
		Employee e = staffRepository.findById(id).orElse(null);
		if (e == null) return;
		orderManagement.findAll(Pageable.unpaged())
			.stream()
			.forEach(o -> {
				o.removeEmployee(e);
				orderManagement.save(o);
			});
		staffRepository.deleteById(id);
	}

	/**
	 * Returns a {@link Set} of all {@link Employee}s of given {@link JobType}
	 * available in the full time range between the given start and end
	 * {@link LocalDateTime}s.
	 *
	 * The range is inclusive, e.g., {@link Employee}s part of an order that lasts
	 * until the given start
	 * are not considered available.
	 *
	 * @param job    the {@link JobType} of {@link Employee}s to search for
	 * @param start  a {@link LocalDateTime} describing the start of the range the
	 *               employee is needed for
	 * @param finish a {@link LocalDateTime} describing the end of the range the
	 *               employee is needed for
	 * @return a {@link Set} of {@link Employee}s matching the requirements
	 */
	public Set<Employee> getAvailableStaffByJob(JobType job, LocalDateTime start, LocalDateTime finish) {
		/*
		 * This function first computes all unavailable employees
		 * to then later take the set difference of all employees and the previous
		 * result
		 * to get a set of all employees available in the given range.
		 */
		Set<Employee> unavailable = orderManagement.findAll(Pageable.unpaged()).stream()
				/*
				* The most complex part is getting all orders overlapping the given range.
				* It can be computed with the following:
				* Let (S, F) be the tuple describing the start and finish of any given order O,
				* let (S', F') be the tuple describing the start and finish of the given date range.
				* Then the following holds:
				*
				* @formatter:off
				*   O overlaps with (S', F')
				* ⇔ F ≥ S' ∧ S ≤ F'
				* ⇔ ¬¬(F ≥ S' ∧ S ≤ F')  (¬¬X ⇔ X ∀ X ∈ {0,1})
				* ⇔ ¬(¬(F ≥ S') ∨ ¬(S ≤ F'))  (De Morgan’s laws: ¬(X ∧ Y) ⇔ ¬X ∨ ¬Y ∀ X,Y ∈ {0,1})
				* ⇔ ¬(F < S' ∨ S > F')  (¬(a ≥ b) ⇔ a < b and ¬(a ≤ b) ⇔ a > b ∀ a,b totally ordered)
				* ⇔ ¬(F < S') ∧ ¬(S > F')  (De Morgan’s laws: ¬(X ∨ Y) ⇔ ¬X ∧ ¬Y ∀ X,Y ∈ {0,1})
				* @formatter:on
				*
				* The last expression can be easily implemented with two filters,
				* as chaining filters is equivalent to the logical conjunction,
				* using LocalDateTime::isBefore and LocalDateTime::isAfter,
				* which are < and > on the total ordering of all LocalDateTimes.
				*
				* Sadly, we can’t really avoid lambdas here
				* as function composition and partial application are not first-class citizens in Java.
				*/
				.filter(order -> !order.getFinish().isBefore(start))
				.filter(order -> !order.getStart().isAfter(finish))
				.map(CustomOrder::getStaff)
				.flatMap(Set::stream)
				.filter(employee -> employee.getJob().equals(job))
				.collect(Collectors.toSet());
		return findAll().stream()
				.filter(e -> e.getJob().equals(job))
				.filter(Predicate.not(unavailable::contains))
				.collect(Collectors.toSet());
	}

	public double getWorkingHoursByEmployee(Employee e, YearMonth month) {
		return orderManagement.findAll(Pageable.unpaged()).stream()
				.filter(order -> order.getStaff().contains(e))
				.map(order -> order.getDurationInSecondsDuringMonth(month))
				.collect(Collectors.summingDouble(seconds -> (double) seconds / 3600));
	}
}
