// SPDX-License-Identifier: AGPL-3.0-or-later
// SPDX-FileCopyrightText: 2023-2024 swt23w23
package catering.users;

import org.springframework.ui.Model;
import org.salespointframework.useraccount.Role;
import org.salespointframework.useraccount.UserAccount;
import org.salespointframework.useraccount.web.LoggedIn;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.validation.Errors;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.InitBinder;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;

import jakarta.validation.Valid;

@Controller
public class UserController {

	private final UserManagement userManagement;

	UserController(UserManagement userManagerment){
		this.userManagement = userManagerment;
	};

	// the binder associates the form validator with the corresponding form
	// as a result, it overrides every validator annotation
	@InitBinder("profileForm")
	private void initProfileBinder(WebDataBinder binder) {
		binder.setValidator(new FormValidator<ProfileForm>());
	}

	@InitBinder("registrationForm")
	private void initRegistrationBinder(WebDataBinder binder) {
		binder.setValidator(new FormValidator<RegistrationForm>());
	}

	@InitBinder("adminForm")
	private void initAdminBinder(WebDataBinder binder) {
		binder.setValidator(new FormValidator<AdminForm>());
	}

	@GetMapping("/unauthorized")
	String unauthorized(){
		return "unauthorized";
	}

	@GetMapping("/register")
	String register(Model model) {
		model.addAttribute("registrationForm", RegistrationForm.empty());
		return "register";
	}

	@PostMapping("/register")
	String register(@Valid RegistrationForm form, Errors result) {
		if (
			(!result.hasErrors()) &&
			(!userManagement.getUserByName(form.getUsername()).isEmpty())
		) {
			result.rejectValue("username", "username already in use");
		}
		if (result.hasErrors()){
			return "register";
		}
		userManagement.createCustomer(form.getUsername(),form.getAddress(),form.getPassword(),form.getFullName());
		return "redirect:/login";
	}

	@GetMapping("/login")
	public String loginPage(){
		return "login";
	}

	@GetMapping("/profile")
	@PreAuthorize("isAuthenticated()")
	public String viewProfile(Model model, @LoggedIn UserAccount userAccount) {
		User user = userManagement.getUserByAccount(userAccount).get();
		model.addAttribute("profileForm", new ProfileForm(user.getUsername(), user.getAddress(), user.getFullName(), ""));
		return "profile";
	}

	@PostMapping("/profile")
	@PreAuthorize("isAuthenticated()")
	public String editProfile(@LoggedIn UserAccount userAccount, @Valid ProfileForm form, Errors result, Model model) {
		String redirect = "redirect:/logout";

		if (
			(!result.hasErrors()) &&
			(!userManagement.getUserByName(form.getUsername()).isEmpty()) &&
			(!userManagement.getUserByAccount(userAccount).get().getUsername().equals(form.getUsername()))
		) {
			result.rejectValue("username", "username already in use");
		}
		if (result.hasErrors()){
			return "profile";
		}
		User user = userManagement.getUserByAccount(userAccount).get();

		if (form.getUsername().equals(user.getUsername())) {
			redirect = "redirect:/profile";
		}
		user.setUsername(form.getUsername());
		user.setFullName(form.getFullName());
		user.setAddress(form.getAddress());
		if (!form.getPassword().isEmpty()) {
			userManagement.setPassword(form.getPassword(), user.getUserAccount());
		}
		userManagement.save(user);

		// by default the user gets logged out by salespoint after changing the user name
		return redirect;
	}

	@GetMapping("/profile/disable")
	@PreAuthorize("hasRole('CUSTOMER')")
	public String disableUser(@LoggedIn UserAccount userAccount) {
		userManagement.disableUserAccount(userAccount);
		return "redirect:/logout";
	}

	@GetMapping("/customers")
	@PreAuthorize("hasRole('ADMIN')")
	public String getCustomer(Model model) {
		model.addAttribute("title", "Kundenverwaltung");
		model.addAttribute("customers", userManagement.getUsers().findAll().stream()
				.filter(customer -> customer.getUserAccount().hasRole(Role.of("CUSTOMER"))).toList());
		return "customers";
	}

	@GetMapping("/customers/remove/{id}")
	@PreAuthorize("hasRole('ADMIN')")
	public String removeCustomer(@PathVariable("id") User user) {
		userManagement.disableUserAccount(user.getUserAccount());
		return "redirect:/customers";
	}

	@GetMapping("/customers/edit/{id}")
	@PreAuthorize("hasRole('ADMIN')")
	public String editCustomer(@PathVariable("id") User user, Model model) {
		model.addAttribute("id", user.getId());
		model.addAttribute("adminForm", new AdminForm(user.getUsername(), user.getAddress(), user.getFullName()));
		return "edit-customer";
	}

	@PostMapping("/customers/edit/{id}")
	@PreAuthorize("hasRole('ADMIN')")
	public String updateCustomer(@PathVariable("id") User user, @Valid AdminForm form, Errors result) {
		if (
			!result.hasErrors() &&
			!userManagement.getUserByName(form.getUsername()).isEmpty() &&
			!user.getUsername().equals(form.getUsername())
		) {
			result.rejectValue("username", "username already in use");
		}
		if (result.hasErrors()){
			return "edit-customer";
		}
		user.setUsername(form.getUsername());
		user.setAddress(form.getAddress());
		user.setFullName(form.getFullName());
		userManagement.save(user);
		return "redirect:/customers";
	}
}
