// SPDX-License-Identifier: AGPL-3.0-or-later
// SPDX-FileCopyrightText: 2023 swt23w23
package catering.order;

import catering.staff.JobType;
import catering.staff.Employee;
import catering.staff.StaffManagement;
import org.javamoney.moneta.Money;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.salespointframework.catalog.Product;
import org.salespointframework.quantity.Quantity;
import org.salespointframework.useraccount.UserAccount;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

import java.time.LocalDateTime;

import static org.assertj.core.api.Assertions.*;
import static org.salespointframework.core.Currencies.EURO;

@SpringBootTest
public class OrderUnitTests {
	@Autowired
	StaffManagement staffManagement;
	Employee employee;
	CustomOrder order;


	@BeforeEach
	void setUp() {
		employee = new Employee("Peter Muffin", JobType.COOK, Money.of(10, EURO));
		staffManagement.save(employee);

		order = new CustomOrder(UserAccount.UserAccountIdentifier.of("12345"),
			new CustomCart(OrderType.RENT_A_COOK,
				LocalDateTime.of(2023, 11, 10, 22, 0),
				LocalDateTime.of(2023, 11, 11, 9, 0)));
	}

	@Test
	void dateTimeFormatter() {
		assertThat(order.getFormattedStart()).isEqualTo("10.11.2023, 22:00 Uhr");
		assertThat(order.getFormattedFinish()).isEqualTo("11.11.2023, 09:00 Uhr");
	}

	@Test
	void addEmployeeToOrder() {
		order.addEmployee(employee);

		// test if employee was added
		assertThat(order.getStaff()).hasSize(1);

		// test if ChargeLine is correct
		assertThat(order.getChargeLines().stream().count()).isEqualTo(1);
		assertThat(order.getChargeLines().getTotal()).isEqualTo(Money.of(11*employee.getWage(), "EUR"));
	}

	@Test
	void orderWithStaffAndProducts() {
		order.addOrderLine(new Product("Brötchen (vegan)",
			Money.of(0.6, "EUR")),
			Quantity.of(10));

		order.addEmployee(employee);

		// test if employee was added
		assertThat(order.getStaff()).hasSize(1);
		assertThat(order.getOrderLines().stream().count()).isEqualTo(1);

		// test if ChargeLine is correct
		assertThat(order.getChargeLines().stream().count()).isEqualTo(1);

		// test if costs are correct
		assertThat(order.getChargeLines().getTotal()).isEqualTo(Money.of(11*employee.getWage(), "EUR"));
		assertThat(order.getOrderLines().getTotal()).isEqualTo(Money.of(10.0*0.6, "EUR"));
		assertThat(order.getTotal()).isEqualTo(Money.of(11.0 * employee.getWage() + 10.0 * 0.6, "EUR"));

		// test for duplication
		order.addEmployee(employee);
		assertThat(order.getStaff()).hasSize(1);
		assertThat(order.getOrderLines().stream().count()).isEqualTo(1);
		assertThat(order.getChargeLines().stream().count()).isEqualTo(1);
		assertThat(order.getChargeLines().getTotal()).isEqualTo(Money.of(11*employee.getWage(), "EUR"));
		assertThat(order.getOrderLines().getTotal()).isEqualTo(Money.of(10.0*0.6, "EUR"));
		assertThat(order.getTotal()).isEqualTo(Money.of(11.0 * employee.getWage() + 10.0 * 0.6, "EUR"));
	}
}
