// SPDX-License-Identifier: AGPL-3.0-or-later
// SPDX-FileCopyrightText: 2023-2024 swt23w23
package catering.orderCatalog;

import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.user;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.model;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.redirectedUrl;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.view;

import java.util.HashMap;
import java.util.Map;

import org.javamoney.moneta.Money;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.salespointframework.catalog.Product;
import org.salespointframework.inventory.UniqueInventory;
import org.salespointframework.inventory.UniqueInventoryItem;
import org.salespointframework.quantity.Quantity;
import org.salespointframework.useraccount.Password;
import org.salespointframework.useraccount.Role;
import org.salespointframework.useraccount.UserAccountManagement;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;

import catering.order.OrderType;

@SpringBootTest
@AutoConfigureMockMvc
public class CatalogControllerIntegrationTests {

	@Autowired
	private MockMvc mockMvc;

	@Autowired
	private CatalogController catalogController;

	@Autowired
	UserAccountManagement userAccountManagement;

	@Autowired
	private CustomCatalogEntryRepository catalogEntryRepository;

	@Autowired
	private UniqueInventory<UniqueInventoryItem> inventory;

	@BeforeEach
	void setUp() {
		if (userAccountManagement.findByUsername("anna").isEmpty()) {
			userAccountManagement.create("anna",
					Password.UnencryptedPassword.of("12345"), Role.of("CUSTOMER"));
		}

		if (catalogEntryRepository.findAll().isEmpty()) {
			Map<Product, Quantity> products = new HashMap<>();
			products.put(inventory.findAll().stream().toList().get(0).getProduct(), Quantity.of(1));

			CustomCatalogEntry testCatalogEntry = new CustomCatalogEntry(
					OrderType.RENT_A_COOK,
					products,
					Money.of(3000, "EUR"));
			catalogEntryRepository.save(testCatalogEntry);
		}
	}

	@Test
	void testCustomerViewsCatalog() throws Exception {
		this.mockMvc.perform(get("/catalog").with(user("anna").roles("CUSTOMER")))
				.andExpect(status().isOk())
				.andExpect(view().name("catalog"))
				.andExpect(model().attributeExists("catalogEntries"));
	}

	@Test
	@WithMockUser(username = "admin", roles = "ADMIN")
	void testAddCatalogEntry() throws Exception {
		this.mockMvc.perform(post("/catalog_editor/addCatalogEntry")
				.param("eventType", "MOBILE_BREAKFAST")
				.param("totalCost", "5000"))
				.andExpect(redirectedUrl("/catalog"));
	}

	@Test
	@WithMockUser(username = "admin", roles = "ADMIN")
	void testRemoveCatalogEntry() throws Exception {
		this.mockMvc.perform(post("/catalog/remove")
				.param("catalogEntryID",
						String.valueOf(catalogEntryRepository.findAll()
								.stream().toList()
								.get(0)
								.getId())))
				.andExpect(redirectedUrl("/catalog"));
	}
}
