// SPDX-License-Identifier: AGPL-3.0-or-later
// SPDX-FileCopyrightText: 2023-2024 swt23w23
package catering.users;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.tuple;
import static org.hamcrest.CoreMatchers.containsString;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.model;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.redirectedUrl;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.view;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.salespointframework.useraccount.UserAccountManagement;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.security.test.context.support.WithAnonymousUser;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;

@AutoConfigureMockMvc
@SpringBootTest
class UserControllerIntegrationTests {
	@Autowired
	MockMvc mvc;

	@Autowired
	private UserManagement userManagement;

	@Autowired
	private UserAccountManagement userAccountManagement;

	void createTestUser(String name) {
		if (userAccountManagement.findByUsername(name).isEmpty()) {
			userManagement.createCustomer(name, "Baum Weg", "123",
					name.substring(0, 1).toUpperCase() + name.substring(1) + " Klaus");
		}
	}

	@BeforeEach
	void setup() throws Exception {
		createTestUser("dieter");
		createTestUser("hans");
		createTestUser("tim");
		createTestUser("gina");
		createTestUser("mona");
		createTestUser("john");
		createTestUser("stefan");
		createTestUser("jacob");
		createTestUser("moana");
		createTestUser("paul");
		createTestUser("mark");
		createTestUser("eva");
		createTestUser("sarah");
		createTestUser("paul");
		createTestUser("maria");
		createTestUser("adam");
	}

	@Test
	void unauthorizedTest() throws Exception {
		mvc.perform(get("/unauthorized"))
				.andExpect(status().isOk())
				.andExpect(view().name("unauthorized"));
	}

	@Test
	void loginPageTest() throws Exception {
		mvc.perform(get("/login"))
				.andExpect(status().isOk())
				.andExpect(view().name("login"));
	}

	@Test
	@WithMockUser(username = "john", password = "123")
	void profileLoadsProperly() throws Exception {
		mvc.perform(get("/profile"))
				.andExpect(status().isOk())
				.andExpect(content().string(containsString("john")))
				.andExpect(content().string(containsString("John Klaus")))
				.andExpect(content().string(containsString("Baum Weg")));
	}

	@Test
	@WithMockUser(username = "dieter", password = "123")
	void redirectAfterChangingOwnUsernameWithDifferentUsername() throws Exception {
		MockHttpServletRequestBuilder createMessage = post("/profile")
				.param("username", "diete")
				.param("password", "123")
				.param("fullName", "Dieter")
				.param("address", "Baum Weg");
		mvc.perform(createMessage).andExpect(redirectedUrl("/logout"));
		assertTrue(userAccountManagement.findByUsername("diete").isPresent());
		assertFalse(userAccountManagement.findByUsername("dieter").isPresent());
	}

	@Test
	@WithMockUser(username = "hans", password = "123")
	void profileRefuseNullByte() throws Exception {
		User pre = userManagement.getUserByName("hans").get();

		mvc.perform(post("/profile")
				.param("username", "hansi")
				.param("password", "\0")
				.param("fullName", "Hans Klaus")
				.param("address", "Baum Weg"))
				.andExpect(status().isOk())
				.andExpect(content().string(containsString("Ungültiges Passwort")));

		User post = userManagement.getUserByName("hans").get();

		assertThat(pre).isEqualTo(post);
	}

	@Test
	@WithMockUser(username = "tim", password = "123")
	void allowEmptyPasswordInProfile() throws Exception {
		String streetPre = userManagement.getUserByName("tim").get().getAddress();

		MockHttpServletRequestBuilder createMessage = post("/profile")
				.param("username", "tim")
				.param("password", "")
				.param("fullName", "Tim Klaus")
				.param("address", "Baum Straße");
		mvc.perform(createMessage)
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/profile"));

		String streetPost = userManagement.getUserByName("tim").get().getAddress();

		assertThat(streetPost).isNotEqualTo(streetPre);
	}

	// This test is necessary, because the password in the profile form is not
	// valided by "validationutils"
	@Test
	@WithMockUser(username = "mona", password = "123")
	void refuseNullPasswordInProfile() throws Exception {
		assertTrue(userManagement.getUserByName("mona").isPresent());
		assertFalse(userManagement.getUserByName("mon").isPresent());
		MockHttpServletRequestBuilder createMessage = post("/profile")
				.param("username", "mon")
				// .param("password", null) can be achieved by simply not setting the password
				// parameter
				.param("fullName", "Mona Klaus")
				.param("address", "Baum Straße");
		mvc.perform(createMessage)
				.andExpect(status().isOk())
				.andExpect(content().string(containsString("Ungültiges Passwort")));
		assertTrue(userAccountManagement.findByUsername("mona").isPresent());
		assertFalse(userAccountManagement.findByUsername("mon").isPresent());

	}

	@Test
	@WithMockUser(username = "gina", password = "123")
	void updateEverythingInProfile() throws Exception {
		assertTrue(userManagement.getUserByName("gina").isPresent());
		assertFalse(userManagement.getUserByName("giina").isPresent());
		MockHttpServletRequestBuilder createMessage = post("/profile")
				.param("username", "giina")
				.param("password", "gina")
				.param("fullName", "Gina")
				.param("address", "Hotel");
		mvc.perform(createMessage)
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/logout"));
		assertFalse(userAccountManagement.findByUsername("gina").isPresent());
		assertThat(userManagement.getUsers().findAll())
				.extracting("username", "fullName", "address")
				.contains(tuple("giina", "Gina", "Hotel"));
		MockHttpServletRequestBuilder login = post("/login")
				.param("username", "giina")
				.param("password", "gina");
		mvc.perform(login)
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/"));
	}

	@Test
	@WithMockUser(username = "stefan", password = "123")
	void updateEverythingInProfileWrong() throws Exception {
		assertTrue(userManagement.getUserByName("stefan").isPresent());
		mvc.perform(post("/profile"))
				.andExpect(status().isOk())
				.andExpect(content().string(containsString("Ungültiger Nutzername")))
				.andExpect(content().string(containsString("Ungültiges Passwort")))
				.andExpect(content().string(containsString("Ungültiger Name")))
				.andExpect(content().string(containsString("Ungültige Addresse")));
		assertFalse(userAccountManagement.findByUsername("steafan").isPresent());
		assertThat(userManagement.getUsers().findAll())
				.extracting("username", "fullName", "address")
				.contains(tuple("stefan", "Stefan Klaus", "Baum Weg"));
	}

	@Test
	@WithAnonymousUser
	void refuseEmptyPasswordDuringRegistration() throws Exception {
		MockHttpServletRequestBuilder createEmptyCustomer = post("/register")
				.param("username", "emptyCustomer")
				.param("password", "")
				.param("fullName", "Dieter")
				.param("address", "Baum Weg");
		mvc.perform(createEmptyCustomer)
				.andExpect(status().isOk())
				.andExpect(content().string(containsString("Ungültiges Passwort")));
		assertFalse(userAccountManagement.findByUsername("emptyCustomer").isPresent());
	}

	@Test
	@WithAnonymousUser
	void createCustomer() throws Exception {
		assertFalse(userAccountManagement.findByUsername("tom").isPresent());
		MockHttpServletRequestBuilder createCustomer = post("/register")
				.param("username", "tom")
				.param("password", "123")
				.param("fullName", "Tom Klaus")
				.param("address", "Baum Weg");
		mvc.perform(createCustomer);
		assertTrue(userAccountManagement.findByUsername("tom").isPresent());
		assertThat(userManagement.getUsers().findAll())
				.extracting("username", "fullName", "address")
				.contains(tuple("tom", "Tom Klaus", "Baum Weg"));
	}

	@Test
	@WithAnonymousUser
	void registrationRefuseNullByte() throws Exception {
		MockHttpServletRequestBuilder createNullCustomer = post("/register")
				.param("username", "didi")
				.param("password", "\0")
				.param("fullName", "Dieter")
				.param("address", "Baum Weg");
		mvc.perform(createNullCustomer)
				.andExpect(status().isOk())
				.andExpect(content().string(containsString("Ungültiges Passwort")));
		assertFalse(userAccountManagement.findByUsername("didi").isPresent());
	}

	@Test
	@WithAnonymousUser
	void registrationFormLoadsProperly() throws Exception {
		mvc.perform(get("/register"))
				.andExpect(status().isOk());
	}

	@Test
	@WithAnonymousUser
	void registerTotallyWrong() throws Exception {
		mvc.perform(post("/register"))
				.andExpect(status().isOk())
				.andExpect(content().string(containsString("Ungültiger Nutzername")))
				.andExpect(content().string(containsString("Ungültiges Passwort")))
				.andExpect(content().string(containsString("Ungültiger Name")))
				.andExpect(content().string(containsString("Ungültige Addresse")));
	}

	@Test
	@WithMockUser(username = "admin", password = "admin", roles = "ADMIN")
	void viewCustomers() throws Exception {
		mvc.perform(get("/customers"))
				.andExpect(status().isOk())
				.andExpect(view().name("customers"))
				.andExpect(model().attributeExists("customers"));
	}

	@Test
	@WithMockUser(username = "admin", password = "admin", roles = "ADMIN")
	void editCustomerDetails() throws Exception {
		assertThat(userManagement.getUsers().findAll())
				.extracting("username", "fullName", "address")
				.contains(tuple("moana", "Moana Klaus", "Baum Weg"));
		mvc.perform(post("/customers/edit/" + userManagement.getUserByName("moana").get().getId())
				.param("username", "moan")
				.param("fullName", "Moan Klaus")
				.param("address", "Baum Straße"))
				.andExpect(redirectedUrl("/customers"));
		assertThat(userManagement.getUsers().findAll())
				.extracting("username", "fullName", "address")
				.contains(tuple("moan", "Moan Klaus", "Baum Straße"));
		mvc.perform(get("/customers/edit/" + userManagement.getUserByName("moan").get().getId()))
				.andExpect(status().isOk())
				.andExpect(view().name("edit-customer"));
	}

	@Test
	@WithMockUser(username = "admin", roles = "ADMIN")
	void removeCustomerTest() throws Exception {
		assertTrue(userAccountManagement.findByUsername("paul").isPresent());
		mvc.perform(get("/customers/remove/" + userManagement.getUserByName("paul").get().getId()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/customers"));
		assertThat(userAccountManagement.findDisabled())
				.extracting("username")
				.contains("paul");
	}

	@Test
	@WithMockUser(username = "jacob", password = "123", roles = "CUSTOMER")
	void disableUserAccount() throws Exception {
		assertTrue(userAccountManagement.findByUsername("jacob").isPresent());
		mvc.perform(get("/profile/disable"))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/logout"));
		assertThat(userAccountManagement.findDisabled())
				.extracting("username")
				.contains("jacob");
	}

	@Test
	@WithAnonymousUser
	void userAlreadyExistsRegistration() throws Exception {
		MockHttpServletRequestBuilder createCustomer = post("/register")
				.param("username", "admin")
				.param("password", "123")
				.param("fullName", "Tom Klaus")
				.param("address", "Baum Weg");
		mvc.perform(createCustomer).andExpect(content().string(containsString("Ungültiger Nutzername")));
	}

	@Test
	@WithMockUser(username = "mark", password = "123")
	void changeUsernameToExistingUsername() throws Exception {
		assertTrue(userAccountManagement.findByUsername("jacob").isPresent());
		assertTrue(userAccountManagement.findByUsername("mark").isPresent());
		MockHttpServletRequestBuilder createCustomer = post("/profile")
				.param("username", "admin")
				.param("password", "")
				.param("fullName", "Mark Klaus")
				.param("address", "Baum Weg");
		mvc.perform(createCustomer).andExpect(content().string(containsString("Ungültiger Nutzername")));
		createCustomer = post("/profile")
				.param("username", "jacob")
				.param("password", "")
				.param("fullName", "Mark Klaus")
				.param("address", "Baum Weg");
		mvc.perform(createCustomer).andExpect(content().string(containsString("Ungültiger Nutzername")));
		assertTrue(userAccountManagement.findByUsername("mark").isPresent());
	}

	@Test
	@WithMockUser(username = "admin", password = "admin", roles = "ADMIN")
	void changeCustomerUsernameToExistingUsername() throws Exception {
		assertTrue(userAccountManagement.findByUsername("eva").isPresent());
		mvc.perform(post("/customers/edit/" + userManagement.getUserByName("eva").get().getId())
			.param("username", "admin")
			.param("fullName", "Eva Klaus")
			.param("address", "Baum Weg")
			).andExpect(content().string(containsString("Ungültiger Nutzername")));
		assertTrue(userAccountManagement.findByUsername("eva").isPresent());
	}

	@Test
	@WithMockUser(username = "admin", password = "admin", roles = "ADMIN")
	void editCustomerDetailsExeptUsername() throws Exception {
		assertThat(userManagement.getUsers().findAll())
				.extracting("username", "fullName", "address")
				.contains(tuple("sarah", "Sarah Klaus", "Baum Weg"));
		mvc.perform(post("/customers/edit/" + userManagement.getUserByName("sarah").get().getId())
				// this form of construction is necessary,
				// because we belive MockMvc internally uses the provided strings directly
				// this leads to different behavor between tests and browser
				.param("username", new String("sarah"))
				.param("fullName", "Sarah Klaus")
				.param("address", "Baum Straße"))
				.andExpect(redirectedUrl("/customers"));
		assertThat(userManagement.getUsers().findAll())
				.extracting("username", "fullName", "address")
				.contains(tuple("sarah", "Sarah Klaus", "Baum Straße"));
		mvc.perform(get("/customers/edit/" + userManagement.getUserByName("sarah").get().getId()))
				.andExpect(status().isOk())
				.andExpect(view().name("edit-customer"));
	}

	@Test
	@WithMockUser(username = "paul", password = "123")
	void updateEverythingInProfileExceptUsername() throws Exception {
		assertTrue(userManagement.getUserByName("paul").isPresent());
		MockHttpServletRequestBuilder createMessage = post("/profile")
				.param("username", new String("paul")) // explained above
				.param("password", "paul")
				.param("fullName", "Paul")
				.param("address", "Hotel");
		mvc.perform(createMessage)
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/profile"));
		assertThat(userManagement.getUsers().findAll())
				.extracting("username", "fullName", "address")
				.contains(tuple("paul", "Paul", "Hotel"));
		MockHttpServletRequestBuilder login = post("/login")
				.param("username", "paul")
				.param("password", "paul");
		mvc.perform(login)
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/"));
	}

	@Test
	@WithMockUser(username = "maria", password = "123")
	void alreadyExistingNameCheckWillNotBePerformedIfFormHasErrorProfile() throws Exception {
		assertTrue(userManagement.getUserByName("maria").isPresent());
		MockHttpServletRequestBuilder createMessage = post("/profile")
				.param("username", "admin")
				.param("password", "\0")
				.param("address", "Baum Straße");
		assertThat(mvc.perform(createMessage).andReturn().getResponse().getContentAsString())
				.contains("Ungültiges Passwort")
				.contains("Ungültiger Name")
				.doesNotContain("Ungültiger Nutzername");
	}

	@Test
	@WithMockUser(username = "admin", password = "admin", roles = "ADMIN")
	void alreadyExistingNameCheckWillNotBePerformedIfFormHasErrorCustomer() throws Exception {
		assertThat(userManagement.getUsers().findAll())
				.extracting("username", "fullName", "address")
				.contains(tuple("adam", "Adam Klaus", "Baum Weg"));
		assertThat(mvc.perform(post("/customers/edit/" + userManagement.getUserByName("adam").get().getId())
				.param("username", "adam")
				.param("address", "Baum Straße"))
				.andReturn().getResponse().getContentAsString())
				.contains("Ungültiger Name")
				.doesNotContain("Ungültiger Nutzername");
	}

	@Test
	@WithAnonymousUser
	void alreadyExistingNameCheckWillNotBePerformedIfFormHasErrorRegistration() throws Exception {
		MockHttpServletRequestBuilder createCustomer = post("/register")
				.param("username", "admin")
				.param("password", "\0")
				.param("fullName", "Tom Klaus")
				.param("address", "Baum Weg");
		assertThat(mvc.perform(createCustomer).andReturn().getResponse().getContentAsString())
				.contains("Ungültiges Passwort")
				.doesNotContain("Ungültiger Nutzername");
	}

}
